@@ -26,17 +26,11 @@ behaviour as their Pango counterparts.
 
 See L<https://developer.gnome.org/pango/stable/pango-Layout-Objects.html>.
 
-The package uses Text::Layout::FontConfig (included) to organize fonts
-by description.
-
-If module HarfBuzz::Shaper is installed, Text::Layout can use it for
-text shaping.
+The package uses Text::Layout::FontConfig to organize fonts by
+description.
 
 Example, using PDF::API2 integration:
 
-    use PDF::API2;		# or PDF::Builder
-    use Text::Layout;
-
     # Create a PDF document.
     my $pdf = PDF::API2->new;	# or PDF::Builder->new
     $pdf->mediabox( 595, 842 );	# A4, PDF units
@@ -48,9 +42,6 @@ Example, using PDF::API2 integration:
     # Create a markup instance.
     my $layout = Text::Layout->new($pdf);
 
-    # This example uses PDF corefonts only.
-    Text::Layout::FontConfig->register_corefonts;
-
     $layout->set_font_description(Text::Layout::FontConfig->from_string("times 40"));
     $layout->set_markup( q{The <i><span foreground="red">quick</span> <span size="20"><b>brown</b></span></i> fox} );
 
@@ -59,25 +50,24 @@ Example, using PDF::API2 integration:
     $layout->set_alignment("center");
 
     # Render it.
-    $layout->show( 0, 600, $ctx );
-    $pdf->saveas("out.pdf");
+    $layout->show( $x, $y, $ctx );
 
 All PDF::API2 graphic and text methods can still be used, they won't
 interfere with the layout methods.
 
-=head1 NOTES FOR PDF::API2/Builder USERS
+=head1 NOTES FOR PDF::API2 USERS
 
 =head2 Baselines
 
-PDF::API2 and PDF::Builder render texts using the font baseline as origin.
+PDF::API2 renders texts using the font baseline as origin.
 
 This module typesets text in an area of possibly limited width and
 height. The origin is the top left of this area. Currently this area
 contains only a single line of text. This will change in the future
 when line breaking and paragraph formatting is implemented.
 
-PDF::API2 and PDF::Builder coordinates have origin bottom left. This
-module produces information with respect to top left coordinates.
+PDF::API2 coordinates have origin bottom left. This module produces
+information with respect to top left coordinates.
 
 =head1 IMPORTANT NOTES FOR PANGO USERS
 
@@ -91,201 +81,32 @@ PDF::API2 uses the coordinate system as defined in the PDF
 specification. It starts off bottom left. For western text the
 direction is increasing I<x> and B<de>creasing I<y>.
 
-=head1 Pango Conformance Mode
-
-Text::Layout can operate in one of two modes: I<convenience mode>
-(enabled by default), and I<Pango conformance mode>. The desired mode
-can be selected by calling the method set_pango_scaling().
-
 =head2 Pango coordinates
 
 Pango uses two device coordinates units: Pango units and device units.
-Pango units are 1024 (C<PANGO_SCALE>) times the device units.
+Pango units are 1000 (C<PANGO_SCALE>) times the device units.
 
 Several methods have two variants, e.g. get_size() and
 get_pixel_size(). The pixel-variant uses device units while the other
 variant uses Pango units.
 
-In I<convenience mode>, this module assumes no scaling. All units are
-PDF device units (1/72 inch).
+This module assumes no scaling. If you insist on multiplying all
+values by PANGO_SCALE use the set_pango_scale() method to say so and
+we'll divide everything back again internally.
 
 =head2 Pango device units
 
-Device units are used for font rendering. Pango device units are 96dpi
-while PDF uses 72dpi.
+Pango device units are 96dpi while PDF uses 72dpi. This module ignores
+this and uses PDF units everywhere, except for font sizes. Since we
+want e.g. a C<Times 20> font to be of equal size in the two systems,
+it will set the PDF font size to 15.
 
-In I<convenience mode> this is ignored. E.g. a C<Times 20> font 
-will be of equal size in the two systems,
-
-In I<Pango conformance mode> you would need to specify a font size of
-C<15360> to get a 20pt font.
-
-=head1 SUPPORTED MARKUP
-
-Text::Layout recognizes most of the Pango markup as provided by the
-Pango library version 1.50 or newer. However, not everything is supported.
-
-=head2 Span attributes
-
-=over 8
-
-=item font="I<DESC>"   font_desc="I<DESC>"
-
-Specifies a font to be used, e.g. C<Serif 20>.
-
-=item font_face="I<FAM>"   face="I<FAM>"
-
-Specifies a font family to be used.
-
-=item font_family="I<FAM>"
-
-Same as font_face="I<FAM>".
-
-=item size=I<FNUM>   size=I<FNUM>pt   size=I<FNUM>%
-
-Font size in 1024ths of a point (conformance mode), or in points (e.g.
-‘12.5pt’), or a percentage (e.g. ‘200%’), or one of the relative sizes
-‘smaller’ or ‘larger’.
-
-Note that in Pango conformance mode, the actual font size is 96/72
-larger. So C<"45pt"> gives a 60pt font.
-
-=item style="I<STYLE>"   font_style="I<STYLE>"
-
-Specifes the font style, e.g. C<italic>.
-
-=item weight="I<WEIGHT>"   font_weight="I<WEIGHT>"
-
-Specifies the font weight, e.g. C<bold>.
-
-=item foreground="I<COLOR>"   fgcolor="I<COLOR>"   color="I<COLOR>"
-
-Specifies the foreground colour, e.g. C<black>.
-
-=item background="I<COLOR>"   bgcolor="I<COLOR>"
-
-Specifies the background colour, e.g. C<white>.
-
-=item underline="I<TYPE>"
-
-Enables underlining.
-I<TYPE> must be C<none>, C<single>, or C<double>.
-
-=item underline_color="I<COLOR>"
-
-The colour to be used for underlining, if enabled.
-
-=item overline="I<TYPE>"
-
-Enables overlining.
-I<TYPE> must be C<none>, C<single>, or C<double>.
-
-=item overline_color="I<COLOR>"
-
-The colour to be used for ovderlining, if enabled.
-
-=item strikethrough="I<ARG>"
-
-Enables or diables overlining. I<ARG> must be C<true> or C<1> to
-enable, and C<false> or C<0> to disable.
-
-=item strikethrough_color="I<COLOR>"
-
-The colour to be used for striking, if enabled.
-
-=item rise=C<NUM>
-
-In convenience mode, lowers the text by I<NUM>/1024 of the font size.
-May be negative to rise the text.
-
-In Pango conformance mode, rises the text by I<NUM> units from the baseline.
-May be negative to lower the text.
-
-Note: In Pango conformance mode, C<rise> does B<not> accumulate.
- Use C<baseline_shift> instead.
-
-=item rise=C<NUM>pt   rise=C<NUM>%   rise=C<NUM>em   rise=C<NUM>ex
-
-Rises the text from the baseline. May be negative to lower the text.
-
-Units are points if postfixed by B<pt>, and a percentage of the current font size if postfixed by B<%>.
-
-B<em> units are equal to the current font size, B<ex> half the font size.
-
-Note: This is not (yet?) part of the Pango markup standard.
-
-=item baseline_shift=C<NUM>   beseline_shift=C<NUM>pt   baseline_shift=C<NUM>%
-
-Like C<rise>, but accumulates.
-
-=back
-
-Also supported but not part of the official Pango Markup specification.
-
-=over 8
-
-=item href="C<URL>"
-
-Creates a clickable target that activates the I<URL>.
-
-=back
-
-=head2 Shortcuts
-
-Equivalent C<span> attributes for shortcuts.
-
-=over 8
-
-=item b
-
-weight=bold
-
-=item big
-
-larger
-
-=item emp
-
-style=italic
-
-=item i
-
-style=italic
-
-=item s
-
-strikethrough=true
-
-=item small
-
-size=smaller
-
-=item strong
-
-weight=bold
-
-=item sub
-
-size=smaller rise=-30%
-
-=item sup
-
-size=smaller rise=30%
-
-=item tt
-
-face=monospace
-
-=item u
-
-underline=single
-
-=back
+I<DBD: Is this really a good idea?>
 
 =cut
 
 use constant {
-    PANGO_SCALE		=> 1024,
+    PANGO_SCALE		=> 1000,
     PANGO_DEVICE_UNITS	=>   96,
     PDF_DEVICE_UNITS	=>   72,
 };
@@ -302,16 +123,9 @@ use Text::Layout::FontConfig;
 
 =item new( $pdf )
 
-Creates a new layout instance for PDF::API2. This is for convenience.
-It is equivalent to
+Creates a new layout instance.
 
-    use Text::Layout::PDFAPI2;
-    $layout = Text::Layout::PDFAPI2->new($pdf);
-
-For other implementations only the above method can be used.
-
-The argument is the I<context> for text formatting. In the case of
-PDF::API2 this will be a PDF::API2 object.
+The arguments are passed to the backend. Read its documentation.
 
 =back
 
@@ -320,22 +134,33 @@ PDF::API2 this will be a PDF::API2 object.
 sub new {
     my ( $pkg, @data ) = @_;
 
-    $pkg =~ s/::$//;
-    if ( $pkg eq __PACKAGE__ ) {
-	# For convenience (and backward compatibility)...
-	unless ( @data >= 1 && ref($data[0]) =~ /^PDF::(API2|Builder)\b/ ) {
-	    croak("Please instanciate a backend, e.g. ".__PACKAGE__."::PDFAPI2");
-	}
-	require Text::Layout::PDFAPI2;
-	return Text::Layout::PDFAPI2->new(@data); # CUL8R
+    my $be = __PACKAGE__."::Markdown";
+
+    if ( @data >= 2 && $data[0] eq "backend") {
+	shift;
+	$be = shift;
     }
-    bless { _context => undef,
+    if ( $INC{"PDF/API2.pm"} ) {
+	$be = __PACKAGE__."::PDFAPI2";
+    }
+    elsif ( $INC{"PDF/Builder.pm"} ) {
+	$be = __PACKAGE__."::PDFBuilder";
+    }
+    elsif ( $INC{"Cairo.pm"} ) {
+	$be = __PACKAGE__."::Cairo";
+    }
+    elsif ( $INC{"Pango.pm"} ) {
+	$be = __PACKAGE__."::Pango";
+    }
+    eval "require $be" or croak($@);
+
+    bless { _be      => $be,
+	    _context => $be->init(@data),
 	    _fonts   => {},
 	    _content => [],
 	    _px2pu   => \&px2pu,
 	    _pu2px   => \&pu2px,
-	    _pango   => 0,
-	  } => $pkg;
+	  };
 }
 
 =over
@@ -369,7 +194,7 @@ sub copy {
 
 =item get_context
 
-Gets the context of this layout.
+Gets the context (PDF document) of this layout.
 
 =back
 
@@ -421,14 +246,12 @@ layout from the markup as this function does not clear all attributes.
 sub set_text {
     my ( $self, $string ) = @_;
     $self->{_content} =
-      [ { type  => "text",
-	  text  => $string,
+      [ { text  => $string,
 	  font  => $self->{_currentfont},
 	  size  => $self->{_currentsize},
 	  color => $self->{_currentcolor},
 	  base  => 0,
 	} ];
-    delete( $self->{_bbcache} );
 }
 
 =over
@@ -456,8 +279,6 @@ sub get_text {
 
 Returns the number of characters in the text of this layout.
 
-Basically the same as length of get_text().
-
 Returns undef if no text has been set.
 
 =back
@@ -487,37 +308,26 @@ Implementation note: Although all markup is parsed, not all is implemented.
 
 =cut
 
-use Text::ParseWords;
-use constant STEP => 0.83333125;	# Cf. Pango
+use constant STEP => 0.8;	# TODO: Optimal value?
 
 my %magstep =
   ( "xx-small"	=> 1.0*STEP*STEP*STEP,
     "x-small"	=> 1.0*STEP*STEP,
     "small"	=> 1.0*STEP,
-    "smaller"	=> 1.0*STEP,
     "medium"	=> 1.0,
     "large"	=> 1.0/STEP,
-    "larger"	=> 1.0/STEP,
     "x-large"	=> 1.0/(STEP*STEP),
     "xx-large"  => 1.0/(STEP*STEP*STEP),
   );
 
 sub set_markup {
     my ( $self, $string ) = @_;
-    confess("set_markup with UNDEF") unless defined $string;
+
     my @stack;
     my @content;
     my $fcur = $self->{_currentfont};
     my $fcol = $self->{_currentcolor};
     my $fsiz = $self->{_currentsize};
-    my $href;
-    my $undl;
-    my $uncl = $fcol;
-    my $ovrl;
-    my $ovcl = $fcol;
-    my $strk;
-    my $stcl = $fcol;
-    my $bcol;
     my $base = 0;
 
     my $try_size = sub {
@@ -540,151 +350,87 @@ sub set_markup {
     my $span;
     $span = sub {
 	my ( $v ) = @_;
-	# Split the span attributes. Note that shellwords is so kind to
-	# split 'font="xx yy"' as a single word 'font=xx yy'.
-	foreach my $k ( shellwords($v) ) {
+	# Split the span attributes.
+	foreach my $k ( split(' ',$v) ) {
 
 	    # key=value
 	    if ( $k =~ /^([-\w]+)=(.+)$/ ) {
 		my ( $k, $v ) = ( $1, $2 );
 
-		# Ignore case unless required.
-		$v = lc $v unless $k =~ /^(link|href|a)$/;
-
-		# Strip quotes. Shouldn't be necessary now we use shellwords.
-		# $v =~ s/^(["'])(.*)\1$/$2/;
+		# Strip quotes.
+		$v =~ s/^(["'])(.*)\1$/$2/;
 
 		# <span font_desc="Sans 20">
-		if ( $k =~ /^(font|font_desc)$/ ) {
-		    $fcur = Text::Layout::FontConfig->from_string($v);
-		    $fsiz = $fcur->get_size if $fcur->get_size;
+		if ( $k =~ /^(font_desc)$/ ) {
+		    # NYI.
 		}
 
 		# <span face="Sans">
-		elsif ( $k =~ /^(face|font_face|font_family)$/ ) {
-		    $fcur = Text::Layout::FontConfig->find_font( $v,
-								 $fcur->{style},
-								 $fcur->{weight});
+		elsif ( $k =~ /^(face|font_family)$/ ) {
+		    $fcur = Text::Layout::FontConfig->find_font($v);
+		    $self->_load_font($fcur);
 		}
 
 		# <span size=20>
-		elsif ( $k =~ /^(size|font_size)$/ ) {
+		elsif ( $k eq "size" ) {
 		    if ( $try_size->($v) ) {
 			#ok
 		    }
-		    elsif ( $v =~ /^(\d+(?:\.\d+)?)pt$/ ) {
-			$fsiz = $self->{_pango}
-			  ? $1 * PANGO_DEVICE_UNITS / PDF_DEVICE_UNITS
-			  : $1;
-			# warn("fsiz \"$v\" -> $fsiz\n");
-		    }
-		    elsif ( $v =~ /^\d+(?:\.\d+)?$/ ) {
-			$fsiz = $self->{_pango}
-			  ? $v * PANGO_DEVICE_UNITS / PDF_DEVICE_UNITS / PANGO_SCALE
-			  : $v;
-			# warn("fsiz \"$v\" -> $fsiz\n");
-		    }
-		    elsif ( $v =~ /^(\d+(?:\.\d+)?)\%$/ ) {
-			$fsiz *= $1 / 100;
-			# warn("fsiz \"$v\" -> $fsiz\n");
+		    elsif ( $v =~ /\d+(?:\.\d+)?$/ ) {
+			$fsiz = 0 + $v;
 		    }
 		    else {
-			carp("Invalid size: \"$v\"\n");
+			carp("Invalid size: $v\n");
 		    }
 		}
 
 		# <span style="Italic">
-		elsif ( $k =~ /^(style|font_style)$/ ) {
+		elsif ( $k eq "style" ) {
 		    $v = Text::Layout::FontConfig::_norm_style($v);
 		    $fcur = Text::Layout::FontConfig->find_font( $fcur->{family},
 					      $v, $fcur->{weight} );
+		    $self->_load_font($fcur);
 		}
 
 		# <span weight="bold">
-		elsif ( $k =~ /^(weight|font_weight)$/ ) {
+		elsif ( $k eq "weight" ) {
 		    $v = Text::Layout::FontConfig::_norm_weight($v);
 		    $fcur = Text::Layout::FontConfig->find_font( $fcur->{family},
 					      $fcur->{style}, $v );
+		    $self->_load_font($fcur);
 		}
 
 		# <span variant="...">
 		# <span stretch="...">
-		elsif ( $k =~ /^(?:font_)?(variant|stretch)$/ ) {
-		    # ignore for now
-		}
-		elsif ( $k =~ /^(features|background_alpha|alpha)$/ ) {
+		elsif ( $k =~ /^variant|stretch$/ ) {
 		    # ignore for now
 		}
 
 		# <span foreground="red">
-		elsif ( $k =~ /^(foreground|fgcolor|color)$/ ) {
+		# We allow 'color' as an alternative.
+		elsif ( $k =~ /^foreground|color$/ ) {
 		    $fcol = $v;
 		}
 
 		# <span background="red">
-		elsif ( $k =~ /^(background|bgcolor)$/ ) {
-		    $bcol = $v;
+		elsif ( $k eq "background" ) {
+		    # NYI.
 		}
 
 		# <span underline="double">
-		elsif ( $k eq "underline" && $v =~ /^(none|single|double)$/i ) {
-		    $undl = lc $v;
-		}
-		elsif ( $k eq "underline_color" ) {
-		    $uncl = lc $v;
-		}
-
-		# <span overline="double">
-		elsif ( $k eq "overline" && $v =~ /^(none|single|double)$/i ) {
-		    $ovrl = lc $v;
-		}
-		elsif ( $k eq "overline_color" ) {
-		    $ovcl = lc $v;
+		elsif ( $k eq "underline" ) {
+		    # NYI.
 		}
 
 		# <span rise=324>
 		elsif ( $k eq "rise" ) {
-		    if ( $v =~ /^(-?\d+(?:\.\d*)?)pt$/ ) {
-			$base = -$1;
-		    }
-		    elsif ( !$self->{_pango} && $v =~ /^(-?\d+(?:\.\d*)?)\%$/ ) {
-			$base = -$1 * $fsiz / 100;
-		    }
-		    elsif ( !$self->{_pango} && $v =~ /^(-?\d+(?:\.\d*)?)e([mx])$/ ) {
-			$base = $2 eq 'x' ? -$1 * $fsiz / 2 : -$1 * $fsiz;
-		    }
-		    else {
-			$v /= PANGO_SCALE;
-			$base = $self->{_pango} ? -$v : $base + $v * $fsiz;
-		    }
-		}
-
-		# <span baseline_shift=324>
-		# Line rise, but accumulate.
-		elsif ( $k eq "baseline_shift" ) {
-		    if ( $v =~ /^(-?\d+(?:\.\d*)?)pt$/ ) {
-			$base -= $1;
-		    }
-		    elsif ( $v =~ /^(-?\d+(?:\.\d*)?)\%$/ ) {
-			$base -= $1 * $fsiz / 100;
-		    }
-		    elsif ( $v =~ /^(-?\d+(?:\.\d*)?)e([mx])$/ ) {
-			$base = $2 eq 'x' ? $1 * $fsiz / 2 : $1 * $fsiz;
-		    }
-		    else {
-			$base -= $self->{_pango} ? $v : -$v * $fsiz;
-		    }
+		    $base += $v / 1024 * $fsiz;
+		    warn("base=$base\n");
 		}
 
 		# <span strikethrough=false>
-		elsif ( $k eq "strikethrough" && $v =~ /^(true|1)$/i ) {
-		    $strk = 1;
-		}
-		elsif ( $k eq "strikethrough" && $v =~ /^(false|0)$/i ) {
-		    $strk = 0;
-		}
-		elsif ( $k eq "strikethrough_color" ) {
-		    $stcl = $v;
+		elsif ( $k eq "strikethrough" ) {
+		    # NYI.
 		}
 
 		# <span fallback=false>
@@ -696,11 +442,6 @@ sub set_markup {
 		elsif ( $k eq "lang" ) {
 		    # Not supported.
 		}
-
-		# <span href="...">
-		elsif ( $k eq "href" ) {
-		    $href = $v;
-		}
 	    }
 
 	    # <span strikethrough>
@@ -714,57 +455,13 @@ sub set_markup {
 	    }
 
 	    else {
-		carp("Invalid span markup: \"$k\"\n");
+		carp("Invalid span markup: $k\n");
 	    }
 	}
     };
 
-    my $image = sub {
-	my ( $v ) = @_;
-
-	my %img;
-	# Split the attributes.
-	foreach my $k ( shellwords($v) ) {
-
-	    # key=value
-	    if ( $k =~ /^([-\w]+)=(.+)$/ ) {
-		my ( $k, $v ) = ( $1, $2 );
-
-		# Ignore case unless required.
-		$v = lc $v unless $k =~ /^(src|id)$/;
-
-		# Strip quotes. Shouldn't be necessary now we use shellwords.
-		# $v =~ s/^(["'])(.*)\1$/$2/;
-
-		if ( $k eq "src" ) {
-		    $img{src} = $v;
-		}
-		elsif ( $k eq "id" ) {
-		    $img{id} = $v;
-		}
-		elsif ( $k =~ /^(width|height|w|h)$/ ) {
-		    $img{$k} = $v;
-		}
-		elsif ( $k =~ /^(x|y)$/ ) {
-		    $img{$k} = $v;
-		}
-		elsif ( $k eq "border" ) {
-		    $img{border} = $v;
-		}
-		else {
-		    carp("Invalid image attribute: \"$k\"\n");
-		}
-	    }
-	    else {
-		carp("Invalid image attribute: \"$k\"\n");
-	    }
-	}
-	return \%img;
-    };
-
-
     # Split the string on markup instructions.
- L: foreach my $a ( split( /(<.*?>)/, $string ) ) {
+    foreach my $a ( split( /(<.*?>)/, $string ) ) {
 
 	# Closing markup, e.g. </b> or </span>.
 	if ( $a =~ m;^<\s*/\s*(\w+)(.*)>$; ) {
@@ -773,9 +470,10 @@ sub set_markup {
 		# Check if it is closing the currently pending markup.
 		if ( $stack[-1][0] =~ /^<\s*$k\b/ ) {
 		    # Restore.
-		    ( undef,
-		      $fcur, $fsiz, $fcol, $undl, $uncl, $ovrl, $ovcl,
-		      $bcol, $strk, $stcl, $base, $href ) = @{$stack[-1]};
+		    $fcur = $stack[-1][1];
+		    $fsiz = $stack[-1][2];
+		    $fcol = $stack[-1][3];
+		    $base = $stack[-1][4];
 		    pop(@stack);
 		}
 		else {
@@ -792,15 +490,11 @@ sub set_markup {
 	}
 
 	# Opening markup, e.g. <b> or <span ...>.
-	elsif ( $a =~ m;^<\s*([-\w]+)(.*)?>$; ) {
+	elsif ( $a =~ m;^<\s*([-\w]+)(.*)>$; ) {
 	    my $k = lc $1;
-	    my $v = $2;
-	    my $closed = $v =~ s;\s*/\s*$;;;
-
+	    my $v = lc $2;
 	    # Save.
-	    push( @stack, [ "<$k".lc($v).">",
-			    $fcur, $fsiz, $fcol, $undl, $uncl, $ovrl, $ovcl,
-			    $bcol, $strk, $stcl, $base, $href ] );
+	    push( @stack, [ "<$k$v>", $fcur, $fsiz, $fcol, $base ] );
 
 	    # <b> <strong>
 	    if ( $k =~ /^(b|strong)$/ ) {
@@ -824,12 +518,12 @@ sub set_markup {
 
 	    # <sub>
 	    elsif ( $k eq "sub" ) {
-		$span->("size=smaller rise=-30%");
+		$span->("size=smaller rise=300");
 	    }
 
 	    # <sup>
 	    elsif ( $k eq "sup" ) {
-		$span->("size=smaller rise=30%");
+		$span->("size=smaller rise=-300");
 	    }
 
 	    # <small>
@@ -852,53 +546,20 @@ sub set_markup {
 		$span->($v);
 	    }
 
-	    # <.../>.
-	    elsif ( ( my $p = $self->get_element_handler($k) ) && $closed ) {
-		push( @content, { type => $k,
-				  %{ $p->parse
-				       ( { type			      => $k,
-					   font			      => $fcur,
-					   size			      => $fsiz,
-					   color		      => $fcol,
-					   bgcolor		      => $bcol,
-					   underline		      => $undl,
-					   underline_color	      => $uncl,
-					   overline		      => $ovrl,
-					   overline_color	      => $ovcl,
-					   strikethrough	      => $strk,
-					   strikethrough_color	      => $stcl,
-					   base			      => $base,
-					   href			      => $href,
-					 }, $k, $v ) } } );
-	    }
-
 	    else {
-		carp("Invalid markup: \"$k\"\n");
-	    }
-	    if ( $closed ) {
-		$a = "</$k>";
-		redo L;
+		carp("Invalid markup: $k\n");
 	    }
 	}
 
 	# Text.
 	else {
 	    push( @content,
-		  { type		     => "text",
-		    text		     => $a,
-		    font		     => $fcur,
-		    size		     => $fsiz,
-		    color		     => $fcol,
-		    bgcolor		     => $bcol,
-		    underline		     => $undl,
-		    underline_color	     => $uncl,
-		    overline		     => $ovrl,
-		    overline_color	     => $ovcl,
-		    strikethrough	     => $strk,
-		    strikethrough_color	     => $stcl,
-		    base		     => $base,
-		    href		     => $href,
-		  } ) if defined $a && $a ne '';
+		  { text  => $a,
+		    font  => $fcur,
+		    size  => $fsiz,
+		    color => $fcol,
+		    base  => $base,
+		  } );
 	}
     }
 
@@ -911,7 +572,6 @@ sub set_markup {
 
     # Store content.
     $self->{_content} = \@content;
-    delete( $self->{_bbcache} );
 }
 
 =over
@@ -958,28 +618,23 @@ $description is a Text::Layout::FontConfig object.
 
 sub set_font_description {
     my ( $self, $description ) = @_;
-    my $o = "Text::Layout::FontDescriptor";
+    my $o = "Text::Layout::FontConfig";
     croak("set_font_description requires a $o object (not $description)")
       unless UNIVERSAL::isa( $description, $o );
 
-    $self->{_currentfont}  = $description;
+    $self->_load_font( $description );
+    $self->set_font( $description );
+    $self->set_font_size($description->{size})
+      if $description->{size};
+}
 
-    if ( my $sz = $description->{size} ) {
-	# Font sizes can be in either Pango units, or device units.
-	# Use heuristics.
-	if ( $sz > 2 * PANGO_SCALE ) { # Pango units
-	    $sz /= PANGO_SCALE;
-	}
-	if ( $self->{_pango} ) {
-	    # Pango uses 96dpi while PDF is 72dpi.
-	    $sz *= PANGO_DEVICE_UNITS / PDF_DEVICE_UNITS;
-	}
-	$self->{_currentsize} = $sz;
+sub _load_font {
+    my ( $self, $description ) = @_;
+    my $font;
+    if ( !$description->{font} && ( $font = $description->{load} ) ) {
+	no strict 'refs';
+	"$self->{_be}::load_font"->( $self, $description );
     }
-
-    $self->{_currentcolor} = $description->{color} || "black";
-
-    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1018,7 +673,6 @@ Implementation note: Only alignment is implemented.
 sub set_width {
     my ( $self, $width ) = @_;
     $self->{_width} = $self->{_pu2px}->($width);
-    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1052,7 +706,6 @@ Implementation note: Height restrictions are not yet implemented.
 sub set_height {
     my ( $self, $height ) = @_;
     $self->{_height} = $self->{_pu2px}->($height);
-    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1236,7 +889,6 @@ Not yet implemented.
 sub set_spacing {
     my ( $self, $spacing ) = @_;
     $self->{_currentspacing} = $self->{_pu2px}->($spacing);
-    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1279,7 +931,6 @@ Not yet implemented.
 sub set_line_spacing {
     my ( $self, $factor ) = @_;
     $self->{_currentlinespacing} = $factor;
-    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1369,9 +1020,8 @@ sub set_alignment {
 	$self->{_alignment} = lc $align;
     }
     else {
-	croak("Invalid alignment: \"$align\"");
+	croak("Invalid alignment: $align");
     }
-    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1503,22 +1153,18 @@ Computes the logical and ink extents of the layout.
 
 Logical extents are usually what you want for positioning things.
 
-Return value in scalar context is a hash ref with 4 values:
- C<x>, C<y>, C<width>, and C<height>
-describing the logical extents of the layout.
-In list context an array of two hashrefs is returned.
-The first reflects the ink extents, the second the logical extents.
+Return value is an array (not a ref!) containing two array refs with 4
+values: C<x>, C<y>, C<width>, and C<height>. The first reflects the
+ink extents, the second the logical extents.
 
-In the extents, C<x> will reflect the offset when text is centered or
-right aligned. It will be zero for left aligned text. For right
-aligned text, it will be the width of the layout.
+C<x> will reflect the offset when text is centered or right aligned.
+It will be zero for left aligned text. For right aligned text, it will
+be the width of the layout.
 
 C<y> will reflect the offset when text is centered vertically or
 bottom aligned. It will be zero for top aligned text.
 
-See also get_pixel_extents below.
-
-Implementation note: If the PDF::API support layer cannot calculate ink,
+Implementation note: Since the PDF support layer cannot calculate ink,
 this function returns two identical extents.
 
 =back
@@ -1527,19 +1173,10 @@ this function returns two identical extents.
 
 sub get_extents {
     my ( $self ) = @_;
-    my $need_ink = wantarray;
 
-    my @bb = $self->get_bbox($need_ink);
-    my $res = { bl    => $bb[0],
-		x     => $bb[1], y      => $bb[2],
-		width => $bb[3], height => $bb[4] };
-    my $ink = $res;
-    if ( @bb > 5 ) {
-	$ink = { bl    => $bb[0],
-		 x     => $bb[5], y      => $bb[6],
-		 width => $bb[7], height => $bb[8] };
-    }
-    return $need_ink ? ( $ink, $res ) : $res;
+    my @bb = @{$self->get_bbox};
+    my $res = [ $bb[0], 0, $bb[2], $bb[3]-$bb[1] ];
+    return ( $res, $res );
 }
 
 =over
@@ -1548,15 +1185,6 @@ sub get_extents {
 
 Same as get_extents, but using device units.
 
-The returned values are suitable for (assuming $pdf_text and
-$pdf_gfx are the PDF text and graphics contexts):
-
-    $layout->render( $x, $y, $pdf_text );
-    $box = $layout->get_pixel_extents;
-    $pdf_gfx->translate( $x, $y );
-    $pdf_gfx->rect( @$box{ qw( x y width height ) } );
-    $pdf_gfx->stroke;
-
 =back
 
 =cut
@@ -1564,19 +1192,9 @@ $pdf_gfx are the PDF text and graphics contexts):
 sub get_pixel_extents {
     my ( $self ) = @_;
 
-    my $need_ink = wantarray;
-
-    my @bb = $self->get_pixel_bbox($need_ink);
-    my $res = { bl    => $bb[0],
-		x     => $bb[1], y      => $bb[2],
-		width => $bb[3], height => $bb[4] };
-    my $ink = $res;
-    if ( @bb > 5 ) {
-	$ink = { bl    => $bb[0],
-		 x     => $bb[5], y      => $bb[6],
-		 width => $bb[7], height => $bb[8] };
-    }
-    return $need_ink ? ( $ink, $res ) : $res;
+    my @bb = @{$self->get_pixel_bbox};
+    my $res = [ $bb[0], 0, $bb[2], $bb[3]-$bb[1] ];
+    return ( $res, $res );
 }
 
 =over
@@ -1597,7 +1215,7 @@ returned.
 
 sub get_size {
     my ( $self ) = @_;
-    my $e = $self->get_extents;
+    my $e = $self->get_extents->[1];
     wantarray
       ? return ( $e->{width}, $e->{height} )
       : return { width => $e->{width}, height => $e->{height} };
@@ -1615,7 +1233,7 @@ Same as get_size().
 
 sub get_pixel_size {
     my ( $self ) = @_;
-    my $e = $self->get_pixel_extents;
+    my $e = $self->get_pixel_extents->[1];
     wantarray
       ? return ( $e->{width}, $e->{height} )
       : return { width => $e->{width}, height => $e->{height} };
@@ -1655,7 +1273,7 @@ See get_iter().
 
 sub get_baseline {
     my ( $self ) = @_;
-    return -$self->get_bbox->[0];
+    return -$self->get_bbox->[3];
 }
 
 =head1 METHODS NOT IMPLEMENTED
@@ -1707,7 +1325,6 @@ Sets the size for the current font.
 sub set_font_size {
     my ( $self, $size ) = @_;
     $self->{_currentsize} = $size;
-    delete( $self->{_bbcache} );
 }
 
 =over
@@ -1716,6 +1333,8 @@ sub set_font_size {
 
 Returns the size of the current font.
 
+NOTE: This is not a Pango API method.
+
 =back
 
 =cut
@@ -1727,54 +1346,68 @@ sub get_font_size {
 
 =over
 
+=item set_font( $font )
+
+Designates a font for this layout using an internal font.
+
+=back
+
+=cut
+
+sub set_font {
+    my ( $self, $font ) = @_;
+    $self->{_currentfont}  = $font;
+    $self->{_currentcolor} = $font->{color} || "black";
+}
+
+=over
+
 =item get_bbox
 
-Returns the bounding box of the text, w.r.t. the (top-left) origin.
+Returns the bounding box of the text, w.r.t. the origin.
 
-bb = ( bl, x, y, width, height )
+     +-----------+   ascend
+     |           |
+     |           |
+     |           |
+     |           |
+     |           |
+     | <-width-> |
+     |           |
+     |           |
+     o-----------+   o = origin baseline
+     |           |
+     |           |
+     +-----------+   descend
 
-bb[0] = baseline distance from the top.
+bb = ( left, descend, right, ascend )
 
-bb[1] = displacement from the left, nonzero for centered and right aligned text
+bb[0] = left is nonzero for centered and right aligned text
 
-bb[2] = displacement from the top, usually zero
+bb[1] = descend is a negative value
 
-bb[3] = advancewidth
+bb[2] - bb[0] = advancewidth
 
-bb[4] = height
+bb[2] = layout width for right aligned text
 
-Note that the bounding box will in general be equal to the font
-bounding box except for the advancewidth.
+bb[3] = ascend is a positive value
 
 NOTE: Some fonts do not include accents on capital letters in the ascend.
 
-If an argument is supplied and true, get_bbox() will attempt to
-calculate the ink extents as well, and add these as another set of 4
-elements,
-
-In list context returns the array of values, in scalar context an
-array ref.
-
 =back
 
 =cut
 
 sub get_pixel_bbox {
-    my ( $self, $all ) = @_;
-    my $res;
-    if ( $self->{_bbcache}
-	 && @{ $self->{_bbcache} } == ($all ? 9 : 5) ) {
-	$res = $self->{_bbcache};
-    }
-    else {
-	$res = $self->{_bbcache} = $self->bbox($all);
-    }
+    my ( $self ) = @_;
+    no strict 'refs';
+    my $res = "$self->{_be}::bbox"->( $self );
     wantarray ? @$res : $res;
 }
 
 sub get_bbox {
-    my ( $self, $all ) = @_;
-    my @res = map { $self->{_px2pu}->($_) } ( $self->get_pixel_bbox($all) );
+    my ( $self ) = @_;
+    my @res = map { $self->{_px2pu}->($_) } ( $self->get_pixel_bbox );
     wantarray ? @res : \@res;
 }
 
@@ -1787,7 +1420,7 @@ graphics environment.
 
 Use this instead of Pango::Cairo::show_layout().
 
-For PDF::API2, $text must be an object created by the $page->text method.
+$text must be an object created by PDF $page->text method.
 
 =back
 
@@ -1795,88 +1428,62 @@ For PDF::API2, $text must be an object created by the $page->text method.
 
 sub show {
     my ( $self, @data ) = @_;
-    $self->render( @data );
+    no strict 'refs';
+    "$self->{_be}::render"->( $self, @data );
 }
 
 =over
 
-=item set_pango_mode( $enable )
+=item set_pango_scale( $scale )
 
-Enable/disable Pango conformance mode.
-See L<Pango Conformance Mode>.
+Sets the pango scaling to $scale, which should be 1000.
 
-Returns the internal Pango scaling factor if enabled.
+Set to 1 to cancel scaling, causing all methods to use pixel units
+instead of Pango units. Set to 0 to get the default behaviour.
 
 =back
 
 =cut
 
-sub set_pango_mode {
-    my ( $self, $conformant ) = @_;
-
-    if ( $conformant ) {
-	delete( $self->{_bbcache} ) unless $self->{_pango};
-	$self->{_px2pu} = sub { $_[0] * PANGO_SCALE };
-	$self->{_pu2px} = sub { $_[0] / PANGO_SCALE };
-	return $self->{_pango} = PANGO_SCALE;
+sub set_pango_scale {
+    my ( $self, $scale ) = @_;
+    $scale //= PANGO_SCALE;
+    if ( $scale == 0 ) {
+	$self->{_px2pu} = \&px2pu;
+	$self->{_pu2px} = \&pu2px;
+    }
+    elsif ( $scale == 1 ) {
+	$self->{_px2pu} = sub { $_[0] };
+	$self->{_pu2px} = sub { $_[0] };
+    }
+    else {
+	carp("Pango scale should be set to @{[PANGO_SCALE]} for Pango conformance")
+	  unless $scale == PANGO_SCALE;
+	$self->{_px2pu} = sub { $_[0] * $scale };
+	$self->{_pu2px} = sub { $_[0] / $scale };
     }
-
-    delete( $self->{_bbcache} ) if $self->{_pango};
-    $self->{_px2pu} = \&px2pu;
-    $self->{_pu2px} = \&pu2px;
-    return $self->{_pango} = 0;
 }
 
-# Legacy.
-*set_pango_scale = \&set_pango_mode;
-
-=over
-
-=item get_pango
-
-See L<Pango Conformance Mode>.
-
-Returns the internal Pango scaling factor if conformance mode is
-enabled, otherwise it returns 1 (one).
-
-=back
-
-=cut
-
 sub get_pango_scale {
     my ( $self ) = @_;
-    $self->{_pango} ? PANGO_SCALE : 1;
+    $self->{_px2pu}->(1);
 }
 
 sub nyi {
     croak("Method \"" . (caller(1))[3] . "\" not implemented");
 }
 
-my $aliens;
-sub register_element {
-    my ( $self, $hd, @tags ) = @_;
-    croak("Element handler for \"$tags[0]\" does not play nicely")
-      unless $hd->DOES(qw(Text::Layout::ElementRole));
-    for ( @tags ) {
-	$aliens->{$_} = $hd;
-    }
-}
-
-sub get_element_handler {
-    my ( $self, $tag ) = @_;
-    return unless $tag;
-    $aliens->{$tag};
-}
+=back
 
 =head1 SEE ALSO
 
 Description of the Pango Markup Language:
-L<https://docs.gtk.org/Pango/pango_markup.html#pango-markup>.
+L<https://developer.gnome.org/pygtk/stable/pango-markup-language.html>.
 
-Documentation of the Pango Layout class:
-L<https://docs.gtk.org/Pango/class.Layout.html>.
+Documentation  of the Pango Layout class:
+L<https://developer.gnome.org/pango/stable/pango-Layout-Objects.html>.
 
-L<PDF::API2>, L<PDF::Builder>, L<HarfBuzz::Shaper>, L<Font::TTF>.
+L<PDF::API2>, L<PDF::Builder>, L<Font::TTF>.
 
 =head1 AUTHOR
 
@@ -1894,6 +1501,8 @@ You can find documentation for this module with the perldoc command.
 Please report any bugs or feature requests using the issue tracker on
 GitHub.
 
+=back
+
 =head1 LICENSE
 
 Copyright (C) 2019, Johan Vromans
